from ftw.testbrowser import browsing
from ftw.testing import MockTestCase
from ftw.topics.browser import tree
from ftw.topics.testing import TOPICS_FUNCTIONAL_TESTING
from plone.app.testing import TEST_USER_ID
from plone.app.testing import setRoles
from plone.dexterity.utils import createContentInContainer
from unittest import TestCase
import transaction


class TestHelperFunctions(MockTestCase):

    def setUp(self):
        super(TestHelperFunctions, self).setUp()

        self.brain_mocks = [
            self.create_dummy(Title='Foo',
                              getPath=lambda: '/plone/foo',
                              getURL=lambda: 'http://nohost/plone/foo'),

            self.create_dummy(Title='Bar',
                              getPath=lambda: '/plone/foo/bar',
                              getURL=lambda: 'http://nohost/plone/foo/bar'),

            self.create_dummy(Title='Baz',
                              getPath=lambda: '/plone/foo/baz',
                              getURL=lambda: 'http://nohost/plone/foo/baz')]

    def test_get_brain_data(self):
        self.assertEqual(map(tree.get_brain_data, self.brain_mocks),

                         [{'title': 'Foo',
                           'path': '/plone/foo',
                           'url': 'http://nohost/plone/foo',
                           'children': []},

                          {'title': 'Bar',
                           'path': '/plone/foo/bar',
                           'url': 'http://nohost/plone/foo/bar',
                           'children': []},

                          {'title': 'Baz',
                           'path': '/plone/foo/baz',
                           'url': 'http://nohost/plone/foo/baz',
                           'children': []}])

    def test_make_treeish(self):
        data = map(tree.get_brain_data, self.brain_mocks)

        self.maxDiff = None
        self.assertEqual(
            tree.make_treeish(data),

            [{'title': 'Foo',
              'path': '/plone/foo',
              'url': 'http://nohost/plone/foo',

              'children': [

                        {'title': 'Bar',
                         'path': '/plone/foo/bar',
                         'url': 'http://nohost/plone/foo/bar',
                         'children': []},

                        {'title': 'Baz',
                         'path': '/plone/foo/baz',
                         'url': 'http://nohost/plone/foo/baz',
                         'children': []},

                        ]},
             ])


class TestTreeView(TestCase):

    layer = TOPICS_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer['portal']

    @browsing
    def test_visible_topic_levels_on_tree_view(self, browser):
        setRoles(self.portal, TEST_USER_ID, ['Manager'])
        tree = self.create_tree(self.portal)
        first = self.create_topic(tree, u'First Level')
        second = self.create_topic(first, u'Second Level')
        self.create_topic(second, u'Third Level')
        transaction.commit()

        browser.login()
        browser.visit(tree)

        self.assertNotIn(
            'Third Level', self.get_content_links_labels(browser),
            'The tree view should only display the first and the second'
            ' level of the tree, not the third.')

        self.assertEquals(
            ['First Level', 'Second Level'],
            self.get_content_links_labels(browser),
            'The tree view should display the first two tree levels.')

    @browsing
    def test_tree_view_has_two_columns(self, browser):
        setRoles(self.portal, TEST_USER_ID, ['Manager'])
        tree = self.create_tree(self.portal)
        self.create_topic(tree, u'Topic 1')
        self.create_topic(tree, u'Topic 2')
        transaction.commit()

        browser.login()
        browser.visit(tree)

        self.assertEquals(
            [['Topic 1'], ['Topic 2']],
            self.get_content_links_per_column(browser),

            'The tree view should have two columns and arrange the first'
            ' level topics in those two columns.')

    @browsing
    def test_first_level_topics_are_sorted_by_title(self, browser):
        setRoles(self.portal, TEST_USER_ID, ['Manager'])
        tree = self.create_tree(self.portal)
        self.create_topic(tree, u'Topic 3')
        self.create_topic(tree, u'Topic 1')
        self.create_topic(tree, u'Topic 2')
        transaction.commit()

        browser.login()
        browser.visit(tree)

        self.assertEquals(
            ['Topic 1', 'Topic 2', 'Topic 3'],
            self.get_content_links_labels(browser),
            'First level topics should be sorted by title.')

    @browsing
    def test_second_level_topics_are_sorted_by_title(self, browser):
        setRoles(self.portal, TEST_USER_ID, ['Manager'])
        tree = self.create_tree(self.portal)
        parent = self.create_topic(tree, u'Parent')
        self.create_topic(parent, u'Topic 3')
        self.create_topic(parent, u'Topic 1')
        self.create_topic(parent, u'Topic 2')
        transaction.commit()

        browser.login()
        browser.visit(tree)

        self.assertEquals(
            ['Parent', 'Topic 1', 'Topic 2', 'Topic 3'],
            self.get_content_links_labels(browser),
            'Second level topics should be sorted by title.')


    def create_tree(self, parent, title=u'Topics'):
        return createContentInContainer(
            parent, 'ftw.topics.TopicTree', title=title)

    def create_topic(self, parent, title='Topic'):
        return createContentInContainer(parent, 'ftw.topics.Topic',
                                        title=title)

    def get_content_links_labels(self, browser):
        links = browser.css('#content-core a')
        return map(lambda item: item.text, links)

    def get_content_links_per_column(self, browser):
        columns = browser.css('#content-core div.listing-column')
        result = []

        for column in columns:
            links = column.css('a')
            result.append(map(lambda item: item.text, links))

        return result



from ftw.builder.testing import BUILDER_LAYER
from ftw.builder.testing import functional_session_factory
from ftw.builder.testing import set_builder_session_factory
from ftw.testing import IS_PLONE_5
from ftw.testing.layer import ComponentRegistryLayer
from ftw.topics.behavior import ITopicSupportSchema
from ftw.topics.tests.utils import add_behaviors
from plone.app.testing import FunctionalTesting
from plone.app.testing import IntegrationTesting
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import applyProfile
from plone.app.testing import ploneSite
from plone.testing import Layer
from plone.testing import z2
from plone.testing import zca
from plone.testing import zodb
from zope.configuration import xmlconfig
import ftw.topics.tests.builders #noqa


class ZCMLLayer(ComponentRegistryLayer):

    def setUp(self):
        super(ZCMLLayer, self).setUp()
        import ftw.topics
        self.load_zcml_file('tests.zcml', ftw.topics.tests)
        self.load_zcml_file('configure.zcml', ftw.topics)

ZCML_LAYER = ZCMLLayer()


class TopicsLayer(PloneSandboxLayer):

    defaultBases = (PLONE_FIXTURE, BUILDER_LAYER)

    def setUpZope(self, app, configurationContext):
        import z3c.autoinclude
        xmlconfig.file('meta.zcml', z3c.autoinclude,
                       context=configurationContext)
        xmlconfig.string(
            '<configure xmlns="http://namespaces.zope.org/zope">'
            '  <includePlugins package="plone" />'
            '</configure>',
            context=configurationContext)

        import ftw.topics
        xmlconfig.file('functional_tests.zcml', ftw.topics.tests,
                       context=configurationContext)

        z2.installProduct(app, 'ftw.simplelayout')
        z2.installProduct(app, 'Products.DateRecurringIndex')

    def setUpPloneSite(self, portal):
        applyProfile(portal, 'ftw.simplelayout.contenttypes:default')
        applyProfile(portal, 'ftw.topics:default')
        applyProfile(portal, 'plone.restapi:default')


TOPICS_FIXTURE = TopicsLayer()
TOPICS_INTEGRATION_TESTING = IntegrationTesting(
    bases=(TOPICS_FIXTURE, ), name='ftw.topics:integration')
TOPICS_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(TOPICS_FIXTURE, ), name='ftw.topics:functional')


class ExampleContentLayer(Layer):

    defaultBases = (TOPICS_FIXTURE, )

    def setUp(self):
        # Stack the component registry
        self['configurationContext'] = zca.stackConfigurationContext(
            self.get('configurationContext'))

        # Stack the database
        self['zodbDB'] = zodb.stackDemoStorage(
            self.get('zodbDB'), name='SimplelayoutTopicsLayer')

        with ploneSite() as portal:
            applyProfile(portal, 'plone.app.contenttypes:default')

            add_behaviors('Document', 'ftw.topics.behavior.ITopicSupportSchema')
            applyProfile(portal, 'ftw.topics.tests:example')


    def tearDown(self):
        # Zap the stacked ZODB
        self['zodbDB'].close()
        del self['zodbDB']
        # Zap the stacked component registry
        del self['configurationContext']


EXAMPLE_CONTENT_FIXTURE = ExampleContentLayer()
EXAMPLE_CONTENT_DEFAULT_FUNCTIONAL = FunctionalTesting(
    bases=(EXAMPLE_CONTENT_FIXTURE, ),
    name='ftw.topics.examplecontent:default:functional')



from ftw.topics import _
from ftw.topics.interfaces import ITopic
from plone.dexterity.content import Container
from plone.directives.form import Schema
from zope import schema
from zope.interface import implements


class ITopicSchema(Schema):

    show_backrefs = schema.Bool(
        title=_(u"Show backreferences"),
        default=True,
        required=False
        )


class Topic(Container):
    implements(ITopic)



from ftw.referencewidget.widget import ReferenceBrowserWidget
from ftw.topics import _
from ftw.topics.interfaces import ITopicRootFinder
from plone.app.relationfield.event import extract_relations
from plone.autoform.interfaces import IFormFieldProvider
from plone.directives import form
from plone.directives.form import widget
from z3c.relationfield.event import _setRelation
from z3c.relationfield.schema import RelationList
from zope.component import getMultiAdapter
from zope.interface import alsoProvides
from zope.interface import Interface


def get_topic_root(widget):
    finder = getMultiAdapter((widget.context, widget.request),
                             ITopicRootFinder)
    return finder.get_topic_root_path()


class ITopicSupportSchema(Interface):
    """Schema interface for the `ITopicSupport` behavior, adding
    a reference field `topics` to the content type.
    """

    widget('topics',
           ReferenceBrowserWidget,
           override=True,
           start=get_topic_root,
           allow_traversal=["ftw.topics.Topic", "ftw.topics.TopicTree"],
           selectable=["ftw.topics.Topic"])
    form.write_permission(topics='ftw.topics.SetTopicReference')
    topics = RelationList(
        title=_(u'label_topics', default=u'Topics'),
        required=False,
    )


alsoProvides(ITopicSupportSchema, IFormFieldProvider)


# From: https://github.com/4teamwork/opengever.core/commit/88e015d
def add_behavior_relations(obj, event):
    """Register relations in behaviors.

    This event handler fixes a bug in plone.app.relationfield, which only
    updates the zc.catalog when an object gets modified, but not when it gets
    added.
    """
    for behavior_interface, name, relation in extract_relations(obj):
        _setRelation(obj, name, relation)



from ftw.upgrade import UpgradeStep


class InstallPloneAppRelationfield(UpgradeStep):
    """Install plone.app.relationfield.
    """

    def __call__(self):
        self.install_upgrade_profile()
        self.ensure_installed('plone.app.relationfield:default')

    def ensure_installed(self, profile):
        portal_setup = self.getToolByName('portal_setup')
        version = portal_setup.getLastVersionForProfile(profile)
        if version not in (None, 'unknown'):
            return

        self.setup_install_profile(profile)



from Acquisition import aq_parent
from Products.CMFPlone.interfaces.siteroot import IPloneSiteRoot
from ftw.topics.interfaces import ITopicRootFinder
from ftw.topics.interfaces import ITopicTree
from plone.app.layout.navigation.interfaces import INavigationRoot
from zope.component import adapts
from zope.component.hooks import getSite
from zope.interface import Interface
from zope.interface import implements


class DefaultTopicTreeFinder(object):
    """The topic tree finder is used for deciding where the topic
    trees are to select from in the "topics" widget.

    It walks up from the current context and returns the next
    INavigationRoot which contains ITopicTree objects,
    or the IPloneSiteRoot if it is reached.
    """

    implements(ITopicRootFinder)
    adapts(Interface, Interface)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def get_topic_root_path(self):
        obj = self.context
        while obj:
            if INavigationRoot.providedBy(obj) and \
                    self._has_direct_topic_trees(obj):
                return '/'.join(obj.getPhysicalPath())

            if IPloneSiteRoot.providedBy(obj):
                return '/'.join(obj.getPhysicalPath())

            obj = aq_parent(obj)

        return '/'.join(getSite().getPhysicalPath())

    def _has_direct_topic_trees(self, context):
        # use contentValues not objectValues because of ftw.trash
        for obj in context.contentValues():
            if ITopicTree.providedBy(obj):
                return True

        return False



from ftw.testing.genericsetup import GenericSetupUninstallMixin
from ftw.testing.genericsetup import apply_generic_setup_layer
from unittest import TestCase


@apply_generic_setup_layer
class TestGenericSetupUninstall(TestCase, GenericSetupUninstallMixin):
    package = 'ftw.topics'



from ftw.testbrowser import browsing
from ftw.topics.testing import EXAMPLE_CONTENT_DEFAULT_FUNCTIONAL
from plone.app.testing import login
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from unittest import TestCase


class TestViewlet(TestCase):

    layer = EXAMPLE_CONTENT_DEFAULT_FUNCTIONAL

    def setUp(self):
        self.portal = self.layer['portal']
        self.request = self.layer['request']

        setRoles(self.portal, TEST_USER_ID, ['Manager'])
        login(self.portal, TEST_USER_NAME)

        self.tree = self.portal.get('topics')

    @browsing
    def test_viewlet_shows_topics(self, browser):
        document = self.portal.get('manufacturing-processes')
        browser.login().visit(document)

        # first link is the topci and the seconds link is the paren object
        self.assertEqual(
            [
                'Manufacturing', 'Topics',
                'Technology', 'Topics'
            ],
            browser.css('#viewlet-below-content-body .topics ul a').text
        )



from plone.app.layout.viewlets.common import ViewletBase
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from ftw.topics.behavior import ITopicSupportSchema


class TopicViewlet(ViewletBase):
    index = ViewPageTemplateFile("templates/viewlet.pt")

    def update(self):
        super(TopicViewlet, self).update()
        self.topics = self.get_topics()

    def get_topics(self):
        topic_support = ITopicSupportSchema(self.context, None)
        if topic_support is None:
            return []

        topic_relations = topic_support.topics

        if not topic_relations:
            return []

        topics = []
        for relation in topic_relations:
            obj = relation.to_object
            if not obj:
                continue

            topics.append(
                {
                    'title': obj.Title(),
                    'url': obj.absolute_url(),
                    'description': obj.Description(),
                    'parent_title': obj.aq_parent.Title(),
                    'parent_url': obj.aq_parent.absolute_url(),
                }
            )
        return topics



from ftw.topics.interfaces import IBackReferenceCollector
from ftw.topics.interfaces import ITopic
from plone.restapi.interfaces import IExpandableElement
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.services import Service
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.interface import implementer
from zope.interface import Interface


@implementer(IExpandableElement)
@adapter(ITopic, Interface)
class TopicBackreferences(object):
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self, expand=False):
        result = {
            "backreferences": {
                "@id": "{}/@backreferences".format(self.context.absolute_url())
            }
        }
        if not expand:
            return result

        collector = getMultiAdapter((self.context, self.request),
                                    IBackReferenceCollector)

        items = []
        for reference in collector._get_brefs_for(self.context):
            item = getMultiAdapter(
                    (reference, self.request), ISerializeToJson
                )()
            items.append(item)

        result["backreferences"]["items"] = items
        return result


class TopicBackreferencesGet(Service):
    def reply(self):
        backreferences = TopicBackreferences(self.context, self.request)
        return backreferences(expand=True)["backreferences"]



from ftw.testbrowser import browsing
from ftw.topics.testing import EXAMPLE_CONTENT_DEFAULT_FUNCTIONAL
from unittest import TestCase
from plone.app.testing import login
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME


class TestTopicRestapi(TestCase):

    layer = EXAMPLE_CONTENT_DEFAULT_FUNCTIONAL

    def setUp(self):
        self.portal = self.layer['portal']
        self.request = self.layer['request']

        setRoles(self.portal, TEST_USER_ID, ['Manager'])
        login(self.portal, TEST_USER_NAME)

        self.tree = self.portal.get('topics')

    @browsing
    def test_backreferences_expandable(self, browser):
        browser.login()
        topic = self.tree.get('manufacturing')
        document = self.portal.get('manufacturing-processes')

        browser.open(document.absolute_url(), method='GET',
                     headers={'Accept': 'application/json'})
        self.assertNotIn('backreferences', browser.json['@components'])

        browser.open(topic.absolute_url(), method='GET',
                     headers={'Accept': 'application/json'})
        self.assertIn('backreferences', browser.json['@components'])

    @browsing
    def test_node_json_representation(self, browser):
        browser.login()
        topic = self.tree.get('manufacturing')
        document = self.portal.get('manufacturing-processes')

        browser.open(topic.absolute_url() + '?expand=backreferences', method='GET',
                     headers={'Accept': 'application/json'})

        self.assertEqual(
            browser.json['@components']['backreferences']['@id'],
            topic.absolute_url() + '/@backreferences',
        )

        self.assertEqual(
            len(browser.json['@components']['backreferences']['items']),
            1,
        )

        document_json = browser.json['@components']['backreferences']['items'][0]

        browser.open(document.absolute_url(), method='GET',
                     headers={'Accept': 'application/json'})

        self.assertEqual(browser.json, document_json)



# pylint: disable=E0211, E0213
# E0211: Method has no argument
# E0213: Method should have "self" as first argument


from plone.app.layout.navigation.interfaces import INavigationRoot
from zope.interface import Interface


class ITopicBrowserLayer(Interface):
    """ftw.topics specific browser layer.
    """


class ITopicTree(Interface):
    """Marker interface for the root object of the topic tree.
    """


class ITopic(Interface):
    """Marker interface for topic objects.
    """


class ITopicRootFinder(Interface):
    """This adapter is used for finding the root of the selectable
    topic trees for the widget.
    Usually all topics in the plone site are referenceable.
    """

    def __init__(context, request):
        """Adapts any object and the request.
        """

    def get_topic_root_path():
        """Returns the path to the topic root.
        A topic root is usually the parent of the ITopicTree.
        """


class ITopicSupport(Interface):
    """DX object providing this interface are extended
    with an additional reference field "topics" for adding
    references on the topic.
    """


class ITopicReferencePresentation(Interface):
    """Representation adapter of backreferences"""

    def __init__(context, request):
        """Adapts context and request"""

    def consume(objects):
        """Stores the objects, which this adapter can display, yields not
        consumed objects"""

    def title():
        """Group title"""

    def position():
        """Order of adapter call"""

    def available():
        """Determines availability"""


class IBackReferenceCollector(Interface):
    """The back reference collector adapter collects the back references
    from the adapted topic object.
    It groups the results into groups.
    The presentation in the topic views is filtered by group.

    The default implementation uses INavigationRoot for grouping.
    """

    def __init__(context, request):
        """The multi-adapter adapts the topic and the request (for easier
        customiziation using browserlayers).
        """

    def __call__(group_by=INavigationRoot):
        """By default the result is grouped by INavigationRoot.
        This means that each object which is providing INavigationRoot may
        act as a group and the content within this object is assigned
        to this group.
        """



from ftw.simplelayout.contenttypes.contents.interfaces import IContentPage
from ftw.testing import MockTestCase
from ftw.topics.browser.contentpage import representation
from ftw.topics.browser.representation import DefaultRepresentation
from ftw.topics.interfaces import ITopicReferencePresentation
from ftw.topics.testing import ZCML_LAYER
from zope.component import getMultiAdapter
from zope.component import queryMultiAdapter
from zope.interface import directlyProvides
from zope.interface.verify import verifyClass


class TestDefaultReferenceRepresentation(MockTestCase):

    layer = ZCML_LAYER

    def setUp(self):
        super(TestDefaultReferenceRepresentation, self).setUp()

        self.request = self.stub_request()
        self.context = self.create_dummy()

    def test_implements_interface(self):
        self.assertTrue(
            ITopicReferencePresentation.implementedBy(DefaultRepresentation))
        verifyClass(ITopicReferencePresentation, DefaultRepresentation)

    def test_component_registered(self):
        self.assertTrue(
            queryMultiAdapter(
                (self.context, self.request),
                ITopicReferencePresentation, name="default_representation"),
            'DefaultRepresentation is not registered correctly.')

    def test_consum(self):
        objects = [self.create_dummy(), self.create_dummy()]

        adapter = getMultiAdapter(
            (self.context, self.request),
            ITopicReferencePresentation, name="default_representation")
        # Default representation adapter consume everything
        adapter.consume(objects)

        self.assertEquals(len(adapter.objects),
                          len(objects))

    def test_render(self):
        objects = [
            self.create_dummy(absolute_url=lambda: '/path1',
                              title_or_id='Title 1'),
            self.create_dummy(absolute_url=lambda: '/path2',
                              title_or_id='Title 2')]

        adapter = getMultiAdapter(
            (self.context, self.request),
            ITopicReferencePresentation, name="default_representation")
        # Default representation adapter consume everything
        adapter.consume(objects)

        rendered = adapter.render()

        self.assertIn('Title 1', rendered)
        self.assertIn('Title 2', rendered)
        self.assertIn('/path1', rendered)
        self.assertIn('/path2', rendered)


class TestContentPageReferenceRepresentation(MockTestCase):
    layer = ZCML_LAYER

    def setUp(self):
        super(TestContentPageReferenceRepresentation, self).setUp()

        self.request = self.stub_request()
        self.context = self.create_dummy()

    def test_implements_interface(self):
        self.assertTrue(
            ITopicReferencePresentation.implementedBy(
                representation.ContentPageRepresentation))
        verifyClass(ITopicReferencePresentation,
                    representation.ContentPageRepresentation)

    def test_component_registered(self):
        self.assertTrue(
            queryMultiAdapter(
                (self.context, self.request),
                ITopicReferencePresentation,
                name="contentpage_representation"),
            'ContentPageRepresentation is not registered correctly.')

    def test_consum(self):
        objects = [self.create_dummy(),
                   self.create_dummy(),
                   self.stub_interface(IContentPage),
                   self.stub_interface(IContentPage),
                   self.stub_interface(IContentPage)]

        adapter = getMultiAdapter(
            (self.context, self.request),
            ITopicReferencePresentation, name="contentpage_representation")
        # Consume only objects which provides IContentPage
        left = list(adapter.consume(objects))

        self.assertEquals(len(left), 2)

        self.assertEquals(len(adapter.objects),
                          3)

    def test_consum_none(self):
        objects = [self.create_dummy(),
                   self.create_dummy()]

        adapter = getMultiAdapter(
            (self.context, self.request),
            ITopicReferencePresentation, name="contentpage_representation")
        left = list(adapter.consume(objects))

        self.assertEquals(len(left), 2)

        self.assertEquals(len(adapter.objects),
                          0)

    def test_render(self):
        dummy1 = self.create_dummy(absolute_url=lambda: '/path1',
                                   title_or_id='Title 1')
        dummy2 = self.create_dummy(absolute_url=lambda: '/path2',
                                   title_or_id='Title 2')

        objects = [dummy1, dummy2]
        directlyProvides(dummy1, IContentPage)
        directlyProvides(dummy2, IContentPage)

        adapter = getMultiAdapter(
            (self.context, self.request),
            ITopicReferencePresentation, name="contentpage_representation")

        self.assertEquals(list(adapter.consume(objects)), [])

        rendered = adapter.render()

        self.assertIn('Title 1', rendered)
        self.assertIn('Title 2', rendered)
        self.assertIn('/path1', rendered)
        self.assertIn('/path2', rendered)



from Acquisition import aq_inner
from DateTime import DateTime
from Products.CMFCore.utils import getToolByName
from copy import deepcopy
from ftw.topics.interfaces import IBackReferenceCollector
from ftw.topics.interfaces import ITopic
from operator import attrgetter
from plone import api
from plone.app.dexterity.behaviors.metadata import IPublication
from plone.app.layout.navigation.interfaces import INavigationRoot
from plone.dexterity.interfaces import IDexterityContent
from plone.memoize import instance
from zc.relation.interfaces import ICatalog
from zope.component import adapts
from zope.component import getUtility
from zope.component.hooks import getSite
from zope.interface import Interface
from zope.interface import implements
from zope.intid.interfaces import IIntIds


class DefaultCollector(object):
    implements(IBackReferenceCollector)
    adapts(ITopic, Interface)

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.group_by = INavigationRoot
        self.mtool = getToolByName(self.context, 'portal_membership')

    def __call__(self, group_by=INavigationRoot):
        self.group_by = group_by
        return self._get_brefs_per_section()

    @instance.memoize
    def get_sections(self):
        catalog = getToolByName(self.context, 'portal_catalog')
        site = getSite()

        result = [{'label': site.Title(),
                   'path': '/'.join(site.getPhysicalPath()),
                   'UID': 'root',
                   'objects': []}]

        def brain_to_item(brain):
            return {'label': brain.Title,
                    'path': brain.getPath(),
                    'UID': brain.UID,
                    'objects': []}

        brains = catalog(object_provides=self.group_by.__identifier__)
        result.extend(map(brain_to_item, brains))
        self._mark_current_section(result)
        return result

    def _mark_current_section(self, sections):
        sections_by_path = sorted(sections,
                                  key=lambda section: section['path'],
                                  reverse=True)

        current_path = '/'.join(self.context.getPhysicalPath())
        for section in sections_by_path:
            if current_path.startswith(section['path']):
                section['is_current_section'] = True
                return

    def _get_brefs_per_section(self):
        """Returns all back references per section.
        """

        section_paths = sorted(
            map(lambda sec: sec['path'], self.get_sections()),
            key=len, reverse=True)

        mapping = dict(map(lambda path: (path, []), section_paths))
        for obj in self._get_merged_brefs():
            path = '/'.join(obj.getPhysicalPath())
            possible_sec_paths = [spath for spath in section_paths
                                  if path.startswith(spath)]

            # use the longest section path
            spath = sorted(possible_sec_paths, key=len, reverse=True)[0]
            mapping[spath].append(obj)

        result = []
        for section in map(deepcopy, self.get_sections()):
            if mapping[section['path']]:
                section['objects'] = mapping[section['path']]
                result.append(section)

            elif section.get('is_current_section'):
                result.append(section)

        return result

    def _get_merged_brefs(self):
        """Returns all backrefrences to the current context and its
        similar topic objects (see _get_similar_topic_objects docstring).
        """
        unrestricted_objects = reduce(
            list.__add__,
            map(self._get_brefs_for,
                self._get_similar_topic_objects()))

        return filter(
            self._filter_view_permission, unrestricted_objects)

    def _filter_view_permission(self, obj):
        if self.mtool.checkPermission('View', obj):
            return obj

    def _get_similar_topic_objects(self):
        """Returns all objects which have the same relative path to the
        nearest parent providing self.group_by (INavigationRoot by default)
        as the current context.
        The current context is included.
        """
        topic_path = '/'.join(self.context.getPhysicalPath())

        section_paths = sorted(map(lambda item: item.get('path'),
                                   self.get_sections()),
                               key=len, reverse=True)
        section_path = [spath for spath in section_paths
                        if topic_path.startswith(spath)][0]

        # path of context relative to next subsite or site root
        relative_path = topic_path.replace(section_path + '/', '', 1)

        # possibile similar paths
        similar_paths = map(lambda path: '/'.join((path, relative_path)),
                            section_paths)

        # similar brains, including the current context
        query = {'path': {'query': similar_paths,
                          'depth': 0}}
        catalog = getToolByName(self.context, 'portal_catalog')
        return map(lambda brain: brain.getObject(), catalog(query))

    def _get_brefs_for(self, obj):
        objs = self._get_dx_brefs_for(obj)
        return filter(self._filter_inactive_content, objs)

    def _get_dx_brefs_for(self, obj):
        catalog = getUtility(ICatalog)
        obj_intid = getUtility(IIntIds).getId(aq_inner(obj))
        relations = catalog.findRelations({'to_id': obj_intid})
        return map(attrgetter('from_object'), relations)

    def _filter_inactive_content(self, obj):
        effective_date, expiration_date = None, None
        if IDexterityContent.providedBy(obj):
            effective_date, expiration_date = self._get_dx_publication_dates(obj)

        if not effective_date and not expiration_date:
            return True

        now = DateTime()

        if not api.user.has_permission('Access inactive portal content') and expiration_date:
            if now > expiration_date:
                return False

        if not api.user.has_permission('Access future portal content') and effective_date:
            if now < effective_date:
                return False

        return True

    def _get_dx_publication_dates(self, obj):
        publication = IPublication(obj, None)
        if not publication:  # IPublication is not supported
            return None, None

        effective = publication.effective
        expiration = publication.expires

        # convert from datetime to DateTime as used by archetypes
        return DateTime(effective) if effective else None, \
            DateTime(expiration) if expiration else None



from ftw.testbrowser import browsing
from ftw.testing.testcase import Dummy
from ftw.topics.testing import EXAMPLE_CONTENT_DEFAULT_FUNCTIONAL
from plone.app.testing import login
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.browserlayer.layer import mark_layer
from Products.CMFCore.utils import getToolByName
from unittest import TestCase
import transaction


class TestDefaultTopicView(TestCase):

    layer = EXAMPLE_CONTENT_DEFAULT_FUNCTIONAL
    viewname = 'topic_view'

    def setUp(self):
        self.portal = self.layer['portal']
        self.request = self.layer['request']
        self.subsite = self.portal.get('foo').get('subsite')

        setRoles(self.portal, TEST_USER_ID, ['Manager'])
        login(self.portal, TEST_USER_NAME)

        self.tree = self.portal.get('topics')
        self.node = self.tree.get('manufacturing')
        self.subnode = self.node.get('agile-manufacturing')
        self.topic_technology = self.tree.get('technology')

        self.subsite_tree = self.subsite.get('topics')
        self.subsite_node = self.subsite_tree.get('manufacturing')

        transaction.commit()

        mark_layer(None, Dummy(request=self.request))

    def test_default_topic_view(self):
        portal_types = getToolByName(self.portal, 'portal_types')
        fti = portal_types.get('ftw.topics.Topic')
        self.assertEqual(
            fti.default_view, self.viewname,
            'Expected default view to be %s on "ftw.topics.Topic"'
            ' since only the default GS profile was installed' % (
                self.viewname))

    @browsing
    def test_subnode_listed(self, browser):
        browser.open(self.node, view=self.viewname)
        doc = browser.css('h2.subelements-heading').first.text

        self.assertEqual(doc, 'Topics', 'Expected a "Topics" heading')

        # on node (Manufacturing) should be subnode (Agile Manufacturing)
        link_nodes = browser.css('.subelements-listing a')
        self.assertEqual(len(link_nodes), 2,
                         'Found more or less links than expected')

        self.assertEqual(' '.join(link_nodes.text), 'Agile Manufacturing Quality')

    @browsing
    def test_subnode_has_no_children(self, browser):
        browser.open(self.subnode, view=self.viewname)

        self.assertFalse(browser.css('h2.subelements-heading'),
                         'Expected no "Topics" heading')

    @browsing
    def test_default_section_filter_selection(self, browser):
        browser.login().open(self.node, view=self.viewname)

        self.assertEquals(browser.css('.topic-filter li b').first.text,
                          'Plone site',
                          'Expected section "Plone site" to be ' +
                          'selected by default.')

        filter_links = browser.css('.topic-filter li a').first.text
        self.assertIn('Sub Site', filter_links,
                      'Missing awailable section filter "Sub Site"')

        reference_links = browser.css('.topic-reference-listings a').text
        self.assertIn(
            'Manufacturing processes', reference_links,
            'Link "Manufacturing processes" should be shown')

        self.assertNotIn(
            'Theories', reference_links,
            'Link "Theories" should not be shown, it is in'
            ' the section "Sub Site", not "Plone site"')

    @browsing
    def test_other_section_filter(self, browser):
        browser.login().open(self.node, view=self.viewname)
        browser.find('Sub Site').click()

        self.assertEquals(browser.css('.topic-filter li b').first.text,
                          'Sub Site',
                          'Expected section "Sub Site" to be ' +
                          'selected by default.')

        filter_links = browser.css('.topic-filter li a').first.text
        self.assertIn('Plone site', filter_links,
                      'Missing awailable section filter "Plone site"')

        reference_links = browser.css('.topic-reference-listings a').text
        self.assertNotIn(
            'Manufacturing processes', reference_links,
            'Link "Manufacturing processes" should not be shown, it is in'
            ' the section "Sub Site", not "Plone site"')

        self.assertIn(
            'Theories', reference_links,
            'Link "Theories" should be shown')

    @browsing
    def test_SUBSITE_default_section_filter_selection(self, browser):
        browser.open(self.subsite_node, view=self.viewname)

        self.assertEquals(browser.css('.topic-filter li b').first.text,
                          'Sub Site',
                          'Expected section "Sub Site" to be ' +
                          'selected by default.')

        filter_links = browser.css('.topic-filter li a').first.text
        self.assertIn('Plone site', filter_links,
                      'Missing awailable section filter "Plone site"')

        reference_links = browser.css('.topic-reference-listings a').text
        self.assertNotIn(
            'Manufacturing processes', reference_links,
            'Link "Manufacturing processes" should not be shown, it is in'
            ' the section "Sub Site", not "Plone site"')

        self.assertIn(
            'Theories', reference_links,
            'Link "Theories" should be shown')

    @browsing
    def test_sections_are_always_shown_when_there_are_subsites_and_brefs(self, browser):
        browser.open(self.topic_technology, view=self.viewname)

        self.assertEquals(
            browser.css('.topic-filter li').first.text, 'Plone site',
            'Only Plone site should be shown as section, because there'
            ' are other sections (subsites) - even when there is only one'
            ' section shown.')

    @browsing
    def test_sections_are_not_shown_when_there_are_subsites_but_no_brefs(self, browser):
        self.topic_quality = self.node.get('quality')
        browser.open(self.topic_quality, view=self.viewname)

        self.assertEquals(browser.css('.topic-filter li'), [],
                          'Expect no section, because there is no content')

    @browsing
    def test_sections_are_shown_when_other_sections_have_brefs(self, browser):
        obj = self.subsite_tree.get('technology')
        browser.open(obj, view=self.viewname)

        self.assertEquals(
            ['Plone site', 'Sub Site'],
            browser.css('.topic-filter li').text,
            'Section filter should be visible because other sections have'
            ' backreferences.')

    @browsing
    def test_no_section_are_shown_when_there_are_no_subsites(self, browser):
        # delete all "subsites", so that we have only one "section",
        # which is the site.
        self.portal.manage_delObjects(['foo', 'empty-subsite'])
        transaction.commit()

        browser.open(self.topic_technology, view=self.viewname)

        self.assertEquals(
            0, len(browser.css('.topic-filter')),
            'Sections should not be shown in a non-subsite setup.')

    @browsing
    def test_backreferences_without_view_permissions_are_not_visible(self, browser):
        # setup content and relation
        folder = self.portal.get('foo')
        browser.login().open(folder, view='edit')
        browser.fill({'Related Items': self.subnode})
        browser.find('form.buttons.save').click()

        browser.open(self.subnode, view='topic_view')
        foo_logged_in = browser.css(
                '.topic-reference-listings li a').first.text

        self.assertEquals('Foo', foo_logged_in)

        # remove permissions
        folder.manage_permission('View', roles=[], acquire=False)
        transaction.commit()

        browser.open(self.subnode, view='topic_view')
        foo_logged_out = browser.css('.topic-reference-listings li a')

        self.assertEquals([], foo_logged_out)

    @browsing
    def test_settings_can_hide_backreferences(self, browser):
        browser.login()

        browser.visit(self.topic_technology)
        self.assertTrue(browser.css('.referenceRepresentationListing'))

        browser.visit(self.topic_technology, view='edit')
        browser.fill({'Show backreferences': False}).submit()

        self.assertFalse(browser.css('.referenceRepresentationListing'))



from ftw.topics.interfaces import ITopicBrowserLayer
from ftw.topics.testing import TOPICS_FUNCTIONAL_TESTING
from plone.browserlayer.utils import registered_layers
from unittest import TestCase


class TestBrowserLayer(TestCase):

    layer = TOPICS_FUNCTIONAL_TESTING

    def test_request_layer_active(self):
        layers = registered_layers()
        self.assertIn(ITopicBrowserLayer, layers)
